#!/usr/bin/env python
# -*- coding: utf-8 -*-
# @Date    : 2019-05-10 14:45:36
# @Author  : Your Name (you@example.org)
# @Link    : http://example.org
# @Version : $Id$

import os
from Bio.Seq import Seq

ME_S7 = "CTGTCTCTTATACACATCTCCGAGCCCACGAGAC"
ME_S7_revc = str(Seq(ME_S7).reverse_complement())
S5_ME = "TCGTCGGCAGCGTCAGATGTGTATAAGAGACAG"
S5_ME_revc = str(Seq(S5_ME).reverse_complement())

FRAGMENT_LEN = 1000		## set the maximun fragment length
VECTOR = 'pelps_cd19bbz' 	## Vector name
MIN_READS = 3

TSS_KEY = "geneHancerGenes"

def get_abs_path(filename):
	code_path = os.path.dirname(os.path.abspath(__file__))
	if os.path.exists(filename):
		return filename
	elif os.path.exists(os.path.join(code_path, filename)):
		return os.path.join(code_path, filename)
	else:
		return filename
		# print("{} not exists".format(filename))

bb_conf = {
"geneHancerGenes":"geneHancer/geneHancerGenesTssAll.hg38.bb",
"geneHancerRegElementsDoubleElite":"geneHancer/geneHancerRegElementsDoubleElite.hg38.bb",
"geneHancerInteractionsAll.v2.hg38":"geneHancer/geneHancerInteractionsAll.v2.hg38.bb",
"geneHancerRegElements":"geneHancer/geneHancerRegElementsAll.hg38.bb",
"geneHancerInteractionsDoubleElite.hg38":"geneHancer/geneHancerInteractionsDoubleElite.hg38.bb",
"geneHancerGenesTssDoubleElite.hg38":"geneHancer/geneHancerGenesTssDoubleElite.hg38.bb",
"geneHancerInteractionsDoubleElite.v2.hg38":"geneHancer/geneHancerInteractionsDoubleElite.v2.hg38.bb",
"geneHancerInteractionsAll.hg38":"geneHancer/geneHancerInteractionsAll.hg38.bb",
"unipSplice":"uniprot/unipSplice.bb",
"unipLocCytopl":"uniprot/unipLocCytopl.bb",
"unipModif":"uniprot/unipModif.bb",
"unipChain":"uniprot/unipChain.bb",
"unipConflict":"uniprot/unipConflict.bb",
"unipOther":"uniprot/unipOther.bb",
"unipLocExtra":"uniprot/unipLocExtra.bb",
"unipDisulfBond":"uniprot/unipDisulfBond.bb",
"unipDomain":"uniprot/unipDomain.bb",
"unipMut":"uniprot/unipMut.bb",
"unipAliTrembl":"uniprot/unipAliTrembl.bb",
"unipRepeat":"uniprot/unipRepeat.bb",
"unipStruct":"uniprot/unipStruct.bb",
"unipAliSwissprot":"uniprot/unipAliSwissprot.bb",
"unipLocSignal":"uniprot/unipLocSignal.bb",
"unipLocTransMemb":"uniprot/unipLocTransMemb.bb",
"crispr":"crispr10K/crispr.bb",
"k50.Unique.Mappability":"hoffmanMappability/k50.Unique.Mappability.bb",
"k36.G2A-Converted":"hoffmanMappability/k36.G2A-Converted.bb",
"k24.Unique.Mappability":"hoffmanMappability/k24.Unique.Mappability.bb",
"k36.Unique.Mappability":"hoffmanMappability/k36.Unique.Mappability.bb",
"k24.G2A-Converted":"hoffmanMappability/k24.G2A-Converted.bb",
"k100.G2A-Converted":"hoffmanMappability/k100.G2A-Converted.bb",
"k24.C2T-Converted":"hoffmanMappability/k24.C2T-Converted.bb",
"k50.C2T-Converted":"hoffmanMappability/k50.C2T-Converted.bb",
"k50.G2A-Converted":"hoffmanMappability/k50.G2A-Converted.bb",
"k100.Unique.Mappability":"hoffmanMappability/k100.Unique.Mappability.bb",
"k36.C2T-Converted":"hoffmanMappability/k36.C2T-Converted.bb",
"k100.C2T-Converted":"hoffmanMappability/k100.C2T-Converted.bb",
"tcgaGeneExpr":"tcga/tcgaGeneExpr.bb",
"tcgaTranscExpr":"tcga/tcgaTranscExpr.bb",
"crispr":"crispr/crispr.bb",
"knownGene29":"knownGene29.bb",
"knownGene24":"knownGene24.bb",
"mane.0.5":"mane/mane.0.5.bb",
"miRnaAtlasSample1":"bbi/miRnaAtlasSample1.bb",
"spAnnot":"bbi/uniprot/spAnnot.bb",
"spMut":"bbi/uniprot/spMut.bb",
"spStruct":"bbi/uniprot/spStruct.bb",
"miRnaAtlasSample2":"bbi/miRnaAtlasSample2.bb",
"wgEncodeRegDnaseUwJurkatHotspot.broadPeak":"bbi/wgEncodeRegDnase/wgEncodeRegDnaseUwJurkatHotspot.broadPeak.bb",
"wgEncodeRegDnaseUwBjHotspot.broadPeak":"bbi/wgEncodeRegDnase/wgEncodeRegDnaseUwBjHotspot.broadPeak.bb",
"wgEncodeRegDnaseUwTh1Hotspot.broadPeak":"bbi/wgEncodeRegDnase/wgEncodeRegDnaseUwTh1Hotspot.broadPeak.bb",
"wgEncodeRegDnaseUwHreHotspot.broadPeak":"bbi/wgEncodeRegDnase/wgEncodeRegDnaseUwHreHotspot.broadPeak.bb",
"wgEncodeRegDnaseUwLhcnm2Hotspot.broadPeak":"bbi/wgEncodeRegDnase/wgEncodeRegDnaseUwLhcnm2Hotspot.broadPeak.bb",
"wgEncodeRegDnaseUwHcpepicHotspot.broadPeak":"bbi/wgEncodeRegDnase/wgEncodeRegDnaseUwHcpepicHotspot.broadPeak.bb",
"wgEncodeRegDnaseUwNt2d1Hotspot.broadPeak":"bbi/wgEncodeRegDnase/wgEncodeRegDnaseUwNt2d1Hotspot.broadPeak.bb",
"wgEncodeRegDnaseUwHgfHotspot.broadPeak":"bbi/wgEncodeRegDnase/wgEncodeRegDnaseUwHgfHotspot.broadPeak.bb",
"wgEncodeRegDnaseUwHmvecdblneoHotspot.broadPeak":"bbi/wgEncodeRegDnase/wgEncodeRegDnaseUwHmvecdblneoHotspot.broadPeak.bb",
"wgEncodeRegDnaseUwHvmfHotspot.broadPeak":"bbi/wgEncodeRegDnase/wgEncodeRegDnaseUwHvmfHotspot.broadPeak.bb",
"wgEncodeRegDnaseUwT47dHotspot.broadPeak":"bbi/wgEncodeRegDnase/wgEncodeRegDnaseUwT47dHotspot.broadPeak.bb",
"wgEncodeRegDnaseUwHct116Hotspot.broadPeak":"bbi/wgEncodeRegDnase/wgEncodeRegDnaseUwHct116Hotspot.broadPeak.bb",
"wgEncodeRegDnaseUwRptecHotspot.broadPeak":"bbi/wgEncodeRegDnase/wgEncodeRegDnaseUwRptecHotspot.broadPeak.bb",
"wgEncodeRegDnaseUwHsmmHotspot.broadPeak":"bbi/wgEncodeRegDnase/wgEncodeRegDnaseUwHsmmHotspot.broadPeak.bb",
"wgEncodeRegDnaseUwHbmecHotspot.broadPeak":"bbi/wgEncodeRegDnase/wgEncodeRegDnaseUwHbmecHotspot.broadPeak.bb",
"wgEncodeRegDnaseUwHaepicHotspot.broadPeak":"bbi/wgEncodeRegDnase/wgEncodeRegDnaseUwHaepicHotspot.broadPeak.bb",
"wgEncodeRegDnaseUwHrpepicHotspot.broadPeak":"bbi/wgEncodeRegDnase/wgEncodeRegDnaseUwHrpepicHotspot.broadPeak.bb",
"wgEncodeRegDnaseUwMcf7Estradiolctrl0hrHotspot.broadPeak":"bbi/wgEncodeRegDnase/wgEncodeRegDnaseUwMcf7Estradiolctrl0hrHotspot.broadPeak.bb",
"wgEncodeRegDnaseUwMonocytescd14ro01746Hotspot.broadPeak":"bbi/wgEncodeRegDnase/wgEncodeRegDnaseUwMonocytescd14ro01746Hotspot.broadPeak.bb",
"wgEncodeRegDnaseUwNhdfadHotspot.broadPeak":"bbi/wgEncodeRegDnase/wgEncodeRegDnaseUwNhdfadHotspot.broadPeak.bb",
"wgEncodeRegDnaseUwGm12878Hotspot.broadPeak":"bbi/wgEncodeRegDnase/wgEncodeRegDnaseUwGm12878Hotspot.broadPeak.bb",
"wgEncodeRegDnaseUwAg10803Hotspot.broadPeak":"bbi/wgEncodeRegDnase/wgEncodeRegDnaseUwAg10803Hotspot.broadPeak.bb",
"wgEncodeRegDnaseUwHipepicHotspot.broadPeak":"bbi/wgEncodeRegDnase/wgEncodeRegDnaseUwHipepicHotspot.broadPeak.bb",
"wgEncodeRegDnaseUwPanc1Hotspot.broadPeak":"bbi/wgEncodeRegDnase/wgEncodeRegDnaseUwPanc1Hotspot.broadPeak.bb",
"wgEncodeRegDnaseUwHmfHotspot.broadPeak":"bbi/wgEncodeRegDnase/wgEncodeRegDnaseUwHmfHotspot.broadPeak.bb",
"wgEncodeRegDnaseUwNhdfneoHotspot.broadPeak":"bbi/wgEncodeRegDnase/wgEncodeRegDnaseUwNhdfneoHotspot.broadPeak.bb",
"wgEncodeRegDnaseUwM059jHotspot.broadPeak":"bbi/wgEncodeRegDnase/wgEncodeRegDnaseUwM059jHotspot.broadPeak.bb",
"wgEncodeRegDnaseUwH7hescDiffprota14dHotspot.broadPeak":"bbi/wgEncodeRegDnase/wgEncodeRegDnaseUwH7hescDiffprota14dHotspot.broadPeak.bb",
"wgEncodeRegDnaseUwSkmcHotspot.broadPeak":"bbi/wgEncodeRegDnase/wgEncodeRegDnaseUwSkmcHotspot.broadPeak.bb",
"wgEncodeRegDnaseUwHbvsmcHotspot.broadPeak":"bbi/wgEncodeRegDnase/wgEncodeRegDnaseUwHbvsmcHotspot.broadPeak.bb",
"wgEncodeRegDnaseUwAg09309Hotspot.broadPeak":"bbi/wgEncodeRegDnase/wgEncodeRegDnaseUwAg09309Hotspot.broadPeak.bb",
"wgEncodeRegDnaseUwHrcepicHotspot.broadPeak":"bbi/wgEncodeRegDnase/wgEncodeRegDnaseUwHrcepicHotspot.broadPeak.bb",
"wgEncodeRegDnaseUwHrgecHotspot.broadPeak":"bbi/wgEncodeRegDnase/wgEncodeRegDnaseUwHrgecHotspot.broadPeak.bb",
"wgEncodeRegDnaseUwHuvecHotspot.broadPeak":"bbi/wgEncodeRegDnase/wgEncodeRegDnaseUwHuvecHotspot.broadPeak.bb",
"wgEncodeRegDnaseUwGm04503Hotspot.broadPeak":"bbi/wgEncodeRegDnase/wgEncodeRegDnaseUwGm04503Hotspot.broadPeak.bb",
"wgEncodeRegDnaseUwRpmi7951Hotspot.broadPeak":"bbi/wgEncodeRegDnase/wgEncodeRegDnaseUwRpmi7951Hotspot.broadPeak.bb",
"wgEncodeRegDnaseUwCaco2Hotspot.broadPeak":"bbi/wgEncodeRegDnase/wgEncodeRegDnaseUwCaco2Hotspot.broadPeak.bb",
"wgEncodeRegDnaseUwGm04504Hotspot.broadPeak":"bbi/wgEncodeRegDnase/wgEncodeRegDnaseUwGm04504Hotspot.broadPeak.bb",
"wgEncodeRegDnaseUwLncapHotspot.broadPeak":"bbi/wgEncodeRegDnase/wgEncodeRegDnaseUwLncapHotspot.broadPeak.bb",
"wgEncodeRegDnaseUwMcf7Estradiol100nm1hrHotspot.broadPeak":"bbi/wgEncodeRegDnase/wgEncodeRegDnaseUwMcf7Estradiol100nm1hrHotspot.broadPeak.bb",
"wgEncodeRegDnaseUwHelas3Hotspot.broadPeak":"bbi/wgEncodeRegDnase/wgEncodeRegDnaseUwHelas3Hotspot.broadPeak.bb",
"wgEncodeRegDnaseUwHpfHotspot.broadPeak":"bbi/wgEncodeRegDnase/wgEncodeRegDnaseUwHpfHotspot.broadPeak.bb",
"wgEncodeRegDnaseUwMcf7Hotspot.broadPeak":"bbi/wgEncodeRegDnase/wgEncodeRegDnaseUwMcf7Hotspot.broadPeak.bb",
"wgEncodeRegDnaseUwNhlfHotspot.broadPeak":"bbi/wgEncodeRegDnase/wgEncodeRegDnaseUwNhlfHotspot.broadPeak.bb",
"wgEncodeRegDnaseUwNb4Hotspot.broadPeak":"bbi/wgEncodeRegDnase/wgEncodeRegDnaseUwNb4Hotspot.broadPeak.bb",
"wgEncodeRegDnaseUwHeepicHotspot.broadPeak":"bbi/wgEncodeRegDnase/wgEncodeRegDnaseUwHeepicHotspot.broadPeak.bb",
"wgEncodeRegDnaseUwSaecHotspot.broadPeak":"bbi/wgEncodeRegDnase/wgEncodeRegDnaseUwSaecHotspot.broadPeak.bb",
"wgEncodeRegDnaseUwHsmmtubeHotspot.broadPeak":"bbi/wgEncodeRegDnase/wgEncodeRegDnaseUwHsmmtubeHotspot.broadPeak.bb",
"wgEncodeRegDnaseUwNhaHotspot.broadPeak":"bbi/wgEncodeRegDnase/wgEncodeRegDnaseUwNhaHotspot.broadPeak.bb",
"wgEncodeRegDnaseUwHahHotspot.broadPeak":"bbi/wgEncodeRegDnase/wgEncodeRegDnaseUwHahHotspot.broadPeak.bb",
"wgEncodeRegDnaseUwTh2Hotspot.broadPeak":"bbi/wgEncodeRegDnase/wgEncodeRegDnaseUwTh2Hotspot.broadPeak.bb",
"wgEncodeRegDnaseUwNhberaHotspot.broadPeak":"bbi/wgEncodeRegDnase/wgEncodeRegDnaseUwNhberaHotspot.broadPeak.bb",
"wgEncodeRegDnaseUwAg04450Hotspot.broadPeak":"bbi/wgEncodeRegDnase/wgEncodeRegDnaseUwAg04450Hotspot.broadPeak.bb",
"wgEncodeRegDnaseUwSknmcHotspot.broadPeak":"bbi/wgEncodeRegDnase/wgEncodeRegDnaseUwSknmcHotspot.broadPeak.bb",
"wgEncodeRegDnaseUwHffmycHotspot.broadPeak":"bbi/wgEncodeRegDnase/wgEncodeRegDnaseUwHffmycHotspot.broadPeak.bb",
"wgEncodeRegDnaseUwBonemarrowmscHotspot.broadPeak":"bbi/wgEncodeRegDnase/wgEncodeRegDnaseUwBonemarrowmscHotspot.broadPeak.bb",
"wgEncodeRegDnaseUwHcfHotspot.broadPeak":"bbi/wgEncodeRegDnase/wgEncodeRegDnaseUwHcfHotspot.broadPeak.bb",
"wgEncodeRegDnaseUwCd20ro01778Hotspot.broadPeak":"bbi/wgEncodeRegDnase/wgEncodeRegDnaseUwCd20ro01778Hotspot.broadPeak.bb",
"wgEncodeRegDnaseUwHmveclblHotspot.broadPeak":"bbi/wgEncodeRegDnase/wgEncodeRegDnaseUwHmveclblHotspot.broadPeak.bb",
"wgEncodeRegDnaseUwPrecHotspot.broadPeak":"bbi/wgEncodeRegDnase/wgEncodeRegDnaseUwPrecHotspot.broadPeak.bb",
"wgEncodeRegDnaseUwGm06990Hotspot.broadPeak":"bbi/wgEncodeRegDnase/wgEncodeRegDnaseUwGm06990Hotspot.broadPeak.bb",
"wgEncodeRegDnaseUwHmecHotspot.broadPeak":"bbi/wgEncodeRegDnase/wgEncodeRegDnaseUwHmecHotspot.broadPeak.bb",
"wgEncodeRegDnaseUwK562Hotspot.broadPeak":"bbi/wgEncodeRegDnase/wgEncodeRegDnaseUwK562Hotspot.broadPeak.bb",
"wgEncodeRegDnaseUwAoafHotspot.broadPeak":"bbi/wgEncodeRegDnase/wgEncodeRegDnaseUwAoafHotspot.broadPeak.bb",
"wgEncodeRegDnaseUwHpafHotspot.broadPeak":"bbi/wgEncodeRegDnase/wgEncodeRegDnaseUwHpafHotspot.broadPeak.bb",
"wgEncodeRegDnaseUwAg09319Hotspot.broadPeak":"bbi/wgEncodeRegDnase/wgEncodeRegDnaseUwAg09319Hotspot.broadPeak.bb",
"wgEncodeRegDnaseUwHmvecdbladHotspot.broadPeak":"bbi/wgEncodeRegDnase/wgEncodeRegDnaseUwHmvecdbladHotspot.broadPeak.bb",
"wgEncodeRegDnaseUwHmvecdlyadHotspot.broadPeak":"bbi/wgEncodeRegDnase/wgEncodeRegDnaseUwHmvecdlyadHotspot.broadPeak.bb",
"wgEncodeRegDnaseUwHconfHotspot.broadPeak":"bbi/wgEncodeRegDnase/wgEncodeRegDnaseUwHconfHotspot.broadPeak.bb",
"wgEncodeRegDnaseUwLhcnm2Diff4dHotspot.broadPeak":"bbi/wgEncodeRegDnase/wgEncodeRegDnaseUwLhcnm2Diff4dHotspot.broadPeak.bb",
"wgEncodeRegDnaseUwHmvecdlyneoHotspot.broadPeak":"bbi/wgEncodeRegDnase/wgEncodeRegDnaseUwHmvecdlyneoHotspot.broadPeak.bb",
"wgEncodeRegDnaseUwH7hescDiffprota5dHotspot.broadPeak":"bbi/wgEncodeRegDnase/wgEncodeRegDnaseUwH7hescDiffprota5dHotspot.broadPeak.bb",
"wgEncodeRegDnaseUwHmvecllyHotspot.broadPeak":"bbi/wgEncodeRegDnase/wgEncodeRegDnaseUwHmvecllyHotspot.broadPeak.bb",
"wgEncodeRegDnaseUwHcfaaHotspot.broadPeak":"bbi/wgEncodeRegDnase/wgEncodeRegDnaseUwHcfaaHotspot.broadPeak.bb",
"wgEncodeRegDnaseUwWerirb1Hotspot.broadPeak":"bbi/wgEncodeRegDnase/wgEncodeRegDnaseUwWerirb1Hotspot.broadPeak.bb",
"wgEncodeRegDnaseUwHcmHotspot.broadPeak":"bbi/wgEncodeRegDnase/wgEncodeRegDnaseUwHcmHotspot.broadPeak.bb",
"wgEncodeRegDnaseUwHmvecdadHotspot.broadPeak":"bbi/wgEncodeRegDnase/wgEncodeRegDnaseUwHmvecdadHotspot.broadPeak.bb",
"wgEncodeRegDnaseUwHffHotspot.broadPeak":"bbi/wgEncodeRegDnase/wgEncodeRegDnaseUwHffHotspot.broadPeak.bb",
"wgEncodeRegDnaseUwHepg2Hotspot.broadPeak":"bbi/wgEncodeRegDnase/wgEncodeRegDnaseUwHepg2Hotspot.broadPeak.bb",
"wgEncodeRegDnaseUwGm12865Hotspot.broadPeak":"bbi/wgEncodeRegDnase/wgEncodeRegDnaseUwGm12865Hotspot.broadPeak.bb",
"wgEncodeRegDnaseUwHmvecdneoHotspot.broadPeak":"bbi/wgEncodeRegDnase/wgEncodeRegDnaseUwHmvecdneoHotspot.broadPeak.bb",
"wgEncodeRegDnaseUwHaspHotspot.broadPeak":"bbi/wgEncodeRegDnase/wgEncodeRegDnaseUwHaspHotspot.broadPeak.bb",
"wgEncodeRegDnaseUwHpdlfHotspot.broadPeak":"bbi/wgEncodeRegDnase/wgEncodeRegDnaseUwHpdlfHotspot.broadPeak.bb",
"wgEncodeRegDnaseUwA549Hotspot.broadPeak":"bbi/wgEncodeRegDnase/wgEncodeRegDnaseUwA549Hotspot.broadPeak.bb",
"wgEncodeRegDnaseUwTh1wb54553204Hotspot.broadPeak":"bbi/wgEncodeRegDnase/wgEncodeRegDnaseUwTh1wb54553204Hotspot.broadPeak.bb",
"wgEncodeRegDnaseUwAg04449Hotspot.broadPeak":"bbi/wgEncodeRegDnase/wgEncodeRegDnaseUwAg04449Hotspot.broadPeak.bb",
"wgEncodeRegDnaseUwHacHotspot.broadPeak":"bbi/wgEncodeRegDnase/wgEncodeRegDnaseUwHacHotspot.broadPeak.bb",
"wgEncodeRegDnaseUwWi38Hotspot.broadPeak":"bbi/wgEncodeRegDnase/wgEncodeRegDnaseUwWi38Hotspot.broadPeak.bb",
"wgEncodeRegDnaseUwSknshraHotspot.broadPeak":"bbi/wgEncodeRegDnase/wgEncodeRegDnaseUwSknshraHotspot.broadPeak.bb",
"wgEncodeRegDnaseUwHnpcepicHotspot.broadPeak":"bbi/wgEncodeRegDnase/wgEncodeRegDnaseUwHnpcepicHotspot.broadPeak.bb",
"wgEncodeRegDnaseUwH7hescHotspot.broadPeak":"bbi/wgEncodeRegDnase/wgEncodeRegDnaseUwH7hescHotspot.broadPeak.bb",
"wgEncodeRegDnaseUwNhekHotspot.broadPeak":"bbi/wgEncodeRegDnase/wgEncodeRegDnaseUwNhekHotspot.broadPeak.bb",
"wgEncodeRegDnaseUwHl60Hotspot.broadPeak":"bbi/wgEncodeRegDnase/wgEncodeRegDnaseUwHl60Hotspot.broadPeak.bb",
"wgEncodeRegDnaseUwBe2cHotspot.broadPeak":"bbi/wgEncodeRegDnase/wgEncodeRegDnaseUwBe2cHotspot.broadPeak.bb",
"wgEncodeRegDnaseUwWi384ohtam20nm72hrHotspot.broadPeak":"bbi/wgEncodeRegDnase/wgEncodeRegDnaseUwWi384ohtam20nm72hrHotspot.broadPeak.bb",
"hg38.grcIncidentDb":"bbi/grcIncidentDb/hg38.grcIncidentDb.bb",
"snpediaAll":"bbi/snpediaAll.bb",
"lrg":"bbi/lrg.bb",
"hgmd":"bbi/hgmd.bb",
"interactions":"bbi/interactions.bb",
"clinvarMain":"bbi/clinvar/clinvarMain.bb",
"clinvarCnv":"bbi/clinvar/clinvarCnv.bb",
"lrg":"bbi/lrgRegions/lrg.bb",
"knownGene":"knownGene.bb",
"ncbiRefSeqGenomicDiff":"ncbiRefSeq/ncbiRefSeqGenomicDiff.bb",
"ncbiRefSeqOther":"ncbiRefSeq/ncbiRefSeqOther.bb",
"refSeqFuncElems":"ncbiRefSeq/refSeqFuncElems.bb",
"snp150.bed4":"vai/snp150.bed4.bb",
"snp147.bed4":"vai/snp147.bed4.bb",
"snp151.bed4":"vai/snp151.bed4.bb",
"dbNsfpUniProt":"dbNsfp/dbNsfpUniProt.bb",
"dbNsfpMutationAssessor":"dbNsfp/dbNsfpMutationAssessor.bb",
"dbNsfpSeqChange":"dbNsfp/dbNsfpSeqChange.bb",
"dbNsfpPolyPhen2":"dbNsfp/dbNsfpPolyPhen2.bb",
"dbNsfpLrt":"dbNsfp/dbNsfpLrt.bb",
"dbNsfpMutationTaster":"dbNsfp/dbNsfpMutationTaster.bb",
"dbNsfpVest":"dbNsfp/dbNsfpVest.bb",
"dbNsfpSift":"dbNsfp/dbNsfpSift.bb",
"dbNsfpInterPro":"dbNsfp/dbNsfpInterPro.bb",
"gtexTranscExpr":"gtex/gtexTranscExpr.bb",
}


#!/usr/bin/env python
# -*- coding: utf-8 -*-
# @Date    : 2019-05-10 14:36:28
# @Author  : Your Name (you@example.org)
# @Link    : http://example.org
# @Version : $Id$

import os
import pandas
from collections import Counter, defaultdict

from epiVIA.integration import Integration

class Statistics(object):
	"""docstring for Statistics"""

	def __init__(self, arg):
		super(Statistics, self).__init__()
		self.chimeric_frags = None
		self.chimeric_pairs = None
		self.chimeric_reads = None


def process_integration_site(grouped_integration, gbdb):
	grouped_integration.annotate_TE()
	grouped_integration.annotate_Gene(gbdb)
	# grouped_integration.annotate_Enhancer(gbdb)
	# grouped_integration.nearest_gene()
	return grouped_integration

def ChimeraSummary(chimera_fragments, outdir, gbdb, quality):
	cellCounter = defaultdict(Counter)
	Integration_sites = defaultdict(list)
	Integrated_cells = defaultdict(list)

	try:
		integration_fh = open(os.path.join(outdir, "Integration_sites.list"), 'w')
	except OSError:
		print("Can't open", os.path.join(outdir, "Integration_sites.list"))

	mark_dup = dict()
	for chimera in chimera_fragments:
		barcode = chimera.barcode

		vector_start = chimera.vector_start
		vector_end = chimera.vector_end
		reference_name = chimera.reference_name
		reference_start = chimera.reference_start
		reference_end = chimera.reference_end
		integration = chimera.integration

		mark_dup_key = (barcode, vector_start, vector_end)
		if mark_dup_key in mark_dup:
			if reference_name == mark_dup[mark_dup_key][0] and (mark_dup[mark_dup_key][1] == reference_start or mark_dup[mark_dup_key][2] == reference_end):
				continue
			elif integration.Quality < 30:
				continue
		else:
			mark_dup[mark_dup_key] = (reference_name, reference_start, reference_end)

		cellCounter[barcode]['ChimeraSum'] +=1
		cellCounter[barcode][chimera.chimeric_type] += 1
		
		integration.CellBarcodes = barcode
		integration.ChimeraSource = chimera.chimeric_type
		integration_key = (integration.Chrom, integration.ChrStart, integration.ChrEnd)
		Integration_sites[integration_key].append(integration)
	
	grouped_integration_list = []
	for integration_key in Integration_sites:
		Chrom, ChrStart, ChrEnd = integration_key
		integration_hubs = Integration_sites[integration_key]
		
		barcode_counter = Counter()
		source_counter = Counter()
		Quality = 0
		for integration in integration_hubs:
			barcode_counter[integration.CellBarcodes] += 1
			source_counter[integration.ChimeraSource] += 1
			if Quality < integration.Quality:
				Quality = integration.Quality
		if Quality < quality:
			continue
		VectorStart = integration_hubs[0].VectorStart
		VectorEnd = integration_hubs[0].VectorEnd
		LTR = integration_hubs[0].LTR
		InsertOri = integration_hubs[0].InsertOri
		Genome = integration_hubs[0].Genome

		CellNumber = len(barcode_counter)
		CellBarcodes_tuples = sorted(barcode_counter.items(), key=lambda x:x[1], reverse=True)
		CellBarcodes = [x[0] for x in CellBarcodes_tuples]
		ReadInCell = [x[1] for x in CellBarcodes_tuples]
		ReadNumber = sum(ReadInCell)
		ChimeraSource_tuples = sorted(source_counter.items(), key=lambda x:x[1], reverse=True)
		ChimeraSource = [x[0] for x in ChimeraSource_tuples]
		# ChimeraSource_count = [x[1] for x in ChimeraSource_tuples]
		for barcode in CellBarcodes:
			cellCounter[barcode]['integration'] += 1

		LTR = "None" if LTR == None else LTR
		grouped_integration = Integration(Genome=Genome, Chrom=Chrom, ChrStart=ChrStart, ChrEnd=ChrEnd, VectorStart=VectorStart, VectorEnd=VectorEnd, LTR=LTR, InsertOri=InsertOri, CellBarcodes=CellBarcodes, ReadInCell=ReadInCell, CellNumber=CellNumber, ReadNumber=ReadNumber, ChimeraSource=ChimeraSource, Quality=Quality)
		grouped_integration = process_integration_site(grouped_integration, gbdb)
		grouped_integration_list.append(grouped_integration)

	# p = Pool(n_jobs)
	# for grouped_integration in grouped_integration_list:
	# 	p.apply_async(process_integration_site, (grouped_integration, gbdb))
	# p.close()
	# p.close()

	for grouped_integration in grouped_integration_list:
		if grouped_integration.InsertGene == 'LINC00486':
			continue
		try:
			header
			integration_string, header = grouped_integration.tostring()
			integration_fh.write(integration_string)
		except:
			integration_string, header = grouped_integration.tostring()
			integration_fh.write(header)
			integration_fh.write(integration_string)

		for barcode in grouped_integration.CellBarcodes:
			Integrated_cells[barcode].append(grouped_integration)
	integration_fh.close()

	barcode_integration = defaultdict(dict)
	for barcode in Integrated_cells:
		if len(Integrated_cells[barcode]) == 1:
			integration = Integrated_cells[barcode][0]
			barcode_integration[barcode]['Chrom'] = integration.Chrom
			barcode_integration[barcode]['ChrStart'] = integration.ChrStart
			barcode_integration[barcode]['ChrEnd'] = integration.ChrEnd
			barcode_integration[barcode]['VectorStart'] = integration.VectorStart
			barcode_integration[barcode]['VectorEnd'] = integration.VectorEnd
			barcode_integration[barcode]['InsertGene'] = integration.InsertGene
			barcode_integration[barcode]['ExonIntron'] = integration.ExonIntron
			barcode_integration[barcode]['Enhancer'] = integration.Enhancer
			barcode_integration[barcode]['TEFamily'] = integration.TEFamily
			barcode_integration[barcode]['TEClass'] = integration.TEClass
			barcode_integration[barcode]['Quality'] = integration.Quality
		else:
			i = 0
			for integration in Integrated_cells[barcode]:
				if barcode not in barcode_integration:
					barcode_integration[barcode]['Chrom'] = integration.Chrom
					barcode_integration[barcode]['ChrStart'] = integration.ChrStart
					barcode_integration[barcode]['ChrEnd'] = integration.ChrEnd
					barcode_integration[barcode]['VectorStart'] = integration.VectorStart
					barcode_integration[barcode]['VectorEnd'] = integration.VectorEnd
					barcode_integration[barcode]['InsertGene'] = integration.InsertGene
					barcode_integration[barcode]['ExonIntron'] = integration.ExonIntron
					barcode_integration[barcode]['Enhancer'] = integration.Enhancer
					barcode_integration[barcode]['TEFamily'] = integration.TEFamily
					barcode_integration[barcode]['TEClass'] = integration.TEClass
					barcode_integration[barcode]['Quality'] = integration.Quality			
				elif integration.Chrom == barcode_integration[barcode]['Chrom'] and abs(integration.ChrStart - barcode_integration[barcode]['ChrStart']) <= 1000:
					ChrStart, ChrEnd = sorted([integration.ChrStart, integration.ChrEnd, barcode_integration[barcode]['ChrStart'], barcode_integration[barcode]['ChrEnd']])[1:3]
					barcode_integration[barcode]['ChrStart'] = ChrStart
					barcode_integration[barcode]['ChrEnd'] = ChrEnd
				else:
					i += 1
					barcode = "{}.{}".format(barcode, i)
					barcode_integration[barcode]['Chrom'] = integration.Chrom
					barcode_integration[barcode]['ChrStart'] = integration.ChrStart
					barcode_integration[barcode]['ChrEnd'] = integration.ChrEnd
					barcode_integration[barcode]['VectorStart'] = integration.VectorStart
					barcode_integration[barcode]['VectorEnd'] = integration.VectorEnd
					barcode_integration[barcode]['InsertGene'] = integration.InsertGene
					barcode_integration[barcode]['ExonIntron'] = integration.ExonIntron
					barcode_integration[barcode]['Enhancer'] = integration.Enhancer
					barcode_integration[barcode]['TEFamily'] = integration.TEFamily
					barcode_integration[barcode]['TEClass'] = integration.TEClass
					barcode_integration[barcode]['Quality'] = integration.Quality
	barcode_integration_df = pandas.DataFrame.from_dict(barcode_integration, orient='index', columns=['Chrom', 'ChrStart', 'ChrEnd', 'VectorStart', 'VectorEnd', 'InsertGene', 'ExonIntron', 'Enhancer', 'TEFamily', 'TEClass', 'Quality'])
	barcode_integration_df.index.name = 'CellBarcode'
	barcode_integration_df.to_csv(os.path.join(outdir, "barcode_integration.list"), sep="\t")
	return cellCounter

def VectorSummary(vector_fragments, vector_coverage, outdir):
	cellCounter = defaultdict(Counter)
	Uniq_depth = defaultdict(Counter)
	positions = dict()
	mark_dup = dict()
	for Vectorfrag in vector_fragments:
		barcode = Vectorfrag.barcode
		if Vectorfrag.is_host_alts():
			continue
		mark_dup_key = (Vectorfrag.reference_start, Vectorfrag.reference_end)
		if mark_dup_key in mark_dup:
			continue
		else:
			mark_dup[mark_dup_key] = 1

		cellCounter[barcode]['VectorSum'] += 1

		if Vectorfrag.is_LTR:
			cellCounter[barcode]['LTR'] += 1
			for pos in range(Vectorfrag.reference_start-1, Vectorfrag.reference_end):
				positions[pos] = 1
				Uniq_depth[barcode][pos] += 0.5
			for pos in range(Vectorfrag.alt_start-1, Vectorfrag.alt_end):
				positions[pos] = 1
				Uniq_depth[barcode][pos] += 0.5
		else:
			cellCounter[barcode]['noLTR'] += 1
			for pos in range(Vectorfrag.reference_start-1, Vectorfrag.reference_end):
				positions[pos] = 1
				Uniq_depth[barcode][pos] += 1

	CellBarcodes = cellCounter.keys()
	# csv_path=os.path.join(outdir, "VectorFragment.csv")
	depth_fh = open(vector_coverage, 'w')
	depth_fh.write("barcode,MaxDepth,FragNum,"+",".join([str(x) for x in sorted(positions.keys())]) + "\n")
	for barcode in CellBarcodes:
		max_depth = max(Uniq_depth[barcode].values())
		output = "{},{},{}".format(barcode, max_depth, cellCounter[barcode]['VectorSum'])
		for pos in sorted(positions.keys()):
			output += ",{}".format(Uniq_depth[barcode][pos])
		depth_fh.write(output+"\n")
	return cellCounter


#!/usr/bin/env python
# -*- coding: utf-8 -*-
# @Date    : 2019-04-28 11:12:27
# @Author  : Wenliang Wang (wangwl.me@gmail.com)
# @Link    : http://wangwl.me
# @Version : $0.1$

import os
import sys
import re
import pysam
import pandas as pd
import time
import argparse
from Bio.Seq import Seq
from Bio import SeqIO
from collections import defaultdict
from subprocess import Popen, PIPE

from epiVIA.integration import Integration
from epiVIA.report import ChimeraSummary, VectorSummary

parser = argparse.ArgumentParser(
    description="Find chimera in alignment result with combined host and vector sequence as reference.")
parser.add_argument("bamfile")
parser.add_argument("outdir", help="A directory where the result files (stat on cells and integrations) written")
parser.add_argument("--Provirus", help="Fasta file of provirus sequence that is integrated in the host genome")
parser.add_argument("--Vector", help="Fasta file of vector sequence")
parser.add_argument("--ltr5_start", type=int, default=1, help="Start position of LTR5, required if using vector")
parser.add_argument("--ltr5_end", type=int, required=True, help="End position of LTR5")
parser.add_argument("--ltr3_start", type=int, help="Start position of LTR3, required if using vector")
parser.add_argument("--ltr3_end", type=int, help="End position of LTR3, required if using vector")
parser.add_argument("--HostIndex", required=True, help="The Index prefix of Host genome")
parser.add_argument("--LTRClipLen", default=11, type=int, help="The soft clipped read length to search in LTR")
parser.add_argument("--HostClipLen", default=17, type=int, help="The soft clipped length to search in host genome")
parser.add_argument("--Host2bit", required=True, help="The 2bit (UCSC) file of host genome sequence")
parser.add_argument("--tempdir", help="A directory for temporary files generated in the pipeline")
parser.add_argument("--quality", type=int, default=0, help="Set the MAPQ for the host read that locate the integration site")
parser.add_argument("--candidate_bam", help="A bamfile to write the suspicious reads used for identifying chimeras, saving time if rerun the pipeline")
parser.add_argument("--chimera_bam", help="A bamfile to write the records that are classified as chimeras")
parser.add_argument("--gbdb", default="http://hgdownload.soe.ucsc.edu/gbdb/", help="gbdb directory of UCSC, default: http://hgdownload.soe.ucsc.edu/gbdb/.")
parser.add_argument("--genome", default='hg38', help="Genome name in UCSC, used for annotation of the integration sites")
args = parser.parse_args()

## Illumina sequence adapters

ME_S7 = "CTGTCTCTTATACACATCTCCGAGCCCACGAGAC"
ME_S7_revc = str(Seq(ME_S7).reverse_complement())
S5_ME = "TCGTCGGCAGCGTCAGATGTGTATAAGAGACAG"
S5_ME_revc = str(Seq(S5_ME).reverse_complement())

######################################################   Preflight   ##################################################
if args.Vector:
    try:
        ltr5_start = args.ltr5_start
        ltr5_end = args.ltr5_end
        ltr3_start = args.ltr3_start
        ltr3_end = args.ltr3_end
        ltr_len = ltr5_end - ltr5_start + 1
    except:
        raise Exception("--ltr5_start, --ltr5_end, --ltr3_start, --ltr3_end are required if using Vector sequence.")

    if args.Provirus:
        raise Exception("Only provide provirus or vector sequence")

    try:
        for vector in SeqIO.parse(args.Vector, 'fasta'):
            vector_name = vector.name
            vector_seq = vector.seq
            vector_len = len(vector.seq)
            LTR_SEQ = vector_seq[ltr3_start-1 : ltr3_end]
    except OSError:
        print("Faile to parse the vector fasta file", args.Provirus)

if args.Provirus:
    try:
        args.ltr5_end
    except:
        raise Exception("At least ltr5_end should be provied if using provirus")

    try:
        for vector in SeqIO.parse(args.Provirus, 'fasta'):
            vector_name = vector.name
            vector_seq = vector.seq
            vector_len = len(vector.seq)
            
    except OSError:
        print("Faile to parse the vector fasta file", args.Provirus)
    ltr5_start = 1
    ltr5_end = args.ltr5_end
    ltr_len = ltr5_end
    ltr3_end = vector_len
    ltr3_start = ltr3_end - ltr_len +1
    LTR_SEQ = vector_seq[ltr3_start-1 : ltr3_end]

if not args.Vector and not args.Provirus:
    raise Exception("Vector or Provirus sequence is required")

if args.candidate_bam:
    bamfile_fh = pysam.AlignmentFile(args.bamfile, 'rb')
    candidate_fh = pysam.AlignmentFile(args.candidate_bam, 'wb', template=bamfile_fh)

if args.chimera_bam:
    bamfile_fh = pysam.AlignmentFile(args.bamfile, 'rb')
    chimera_fh = pysam.AlignmentFile(args.chimera_bam, 'wb', template=bamfile_fh)

if not os.path.exists(args.outdir):
    os.makedirs(args.outdir)
if not os.path.exists(args.tempdir):
    os.makedirs(args.tempdir)

####################################################   Enjoy your trip   ##########################################################

# Chimeric fragments: (1) Chimeric Pair; (2) Host Chimeric read; (3) Vector Chimeric read
class ChimericFragment(object):
    """docstring for Chimera"""

    def __init__(self):
        super(ChimericFragment, self).__init__()
        self.integration = Integration()
        self.fragment_length = 0
        self.reference_name = None
        self.reference_start = None
        self.reference_end = None
        self.fragment_seq = []
        self.name = None
        self.barcode = None
        self.chimeric_type = None


class ChimericPair(ChimericFragment):
    """docstring for ChimericPair"""

    def __init__(self, HostRead, VectorRead, integration=None):
        super(ChimericPair, self).__init__()
        # self.HostRead = HostRead
        # self.VectorRead = VectorRead
        self.integration = integration
        self.reference_name = HostRead.reference_name
        self.reference_start = HostRead.reference_start + 1
        self.reference_end = HostRead.reference_end
        self.vector_start = VectorRead.reference_start + 1
        self.vector_end = VectorRead.reference_end
        self.fragment_seq = [HostRead.seq, VectorRead.seq]
        self.readname = HostRead.qname
        self.barcode = get_barcode(HostRead)
        self.chimeric_type = "Pair"

        if self.vector_start >= ltr3_start and self.vector_end <= ltr3_end:
            self.vector_start = self.vector_start - ltr3_start + ltr5_start
            self.vector_end = self.vector_end - ltr3_start + ltr5_start

        if args.chimera_bam:
            chimera_fh.write(HostRead)
            chimera_fh.write(VectorRead)

class ChimericRead(ChimericFragment):
    """docstring for ChimericRead"""

    def __init__(self, read, integration=None):
        super(ChimericRead, self).__init__()
        # self.read = read
        self.reference_name = read.reference_name
        self.reference_start = read.reference_start + 1
        self.reference_end = read.reference_end
        self.vector_start = None,
        self.vector_end = None,
        self.fragment_seq = [read.seq]
        self.integration = integration
        self.readname = read.qname
        self.barcode = get_barcode(read)
        if read.reference_name == vector_name:
            self.chimeric_type = "VectorRead"
        else:
            self.chimeric_type = "HostRead"

        if args.chimera_bam:
            chimera_fh.write(read)


class VectorFragment(object):
    """docstring for VectorFragment"""

    def __init__(self, first, second):
        super(VectorFragment, self).__init__()
        self.reference_start = first.reference_start + 1
        self.reference_end = second.reference_end
        self.alt_start = None
        self.alt_end = None
        self.readname = first.qname
        self.barcode = get_barcode(first)
        self.pair_cigars = [first.cigarstring, second.cigarstring]
        self.first_alts = getAltanatives(first.tags)
        self.second_alts = getAltanatives(second.tags)
        self.first_align_len = first.query_alignment_length
        self.second_align_len = second.query_alignment_length
        self.is_LTR = False

        if self.reference_start >= ltr5_start and self.reference_end <= ltr5_end:
            self.is_LTR = True
            self.alt_start = ltr3_start + self.reference_start - ltr5_start
            self.alt_end = ltr3_start + self.reference_end - ltr5_start
        elif self.reference_start >= ltr3_start and self.reference_end <= ltr3_end:
            self.is_LTR = True
            self.alt_start = self.reference_start
            self.alt_end = self.reference_end
            self.reference_start = self.reference_start - ltr3_start + ltr5_start
            self.reference_end = self.reference_end - ltr3_start + ltr5_start

    def is_host_alts(self):
        first_is_alt = False
        for alt in self.first_alts:
            if alt[0] != vector_name:
                first_is_alt = True
        second_is_alt = False
        for alt in self.second_alts:
            if alt[0] != vector_name:
                second_is_alt = True
        return first_is_alt and second_is_alt


def get_barcode(read):
    # compatible with 10x bam and snaptools bam

    tags_dict = dict(read.tags)
    first_block = read.qname.split(":")[0]
    if 'CB' in tags_dict:
        barcode = tags_dict['CB']
    elif 'CR' in tags_dict:
        barcode = tags_dict['CR']
    elif re.match("^[ATCGN]+$", first_block):
        barcode = first_block
    else:
        barcode = "None"
    return barcode


def getAltanatives(tags):
    tags = dict(tags)
    alt_aligns = []
    if 'XA' in tags:
        alts = tags['XA'].rstrip(";").split(";")
        for alt in alts:
            alt_name, alt_pos, alt_cigar = alt.split(",")[0:3]
            alt_aligns.append([alt_name, alt_pos, alt_cigar])
    return alt_aligns

# @pysnooper.snoop()


def GetClipFrag(read, clipcutoff):
    cigar = read.cigar
    clipped_frag = ''
    candidate_frag = []
    if len(cigar) <= 1:
        return candidate_frag

    if cigar[0][0] == 4 and cigar[0][1] >= clipcutoff:
        clipped_len = cigar[0][1]
        clipped_frag = read.seq[0:clipped_len]
        candidate_frag.append(clipped_frag)

    if cigar[-1][0] == 4 and cigar[-1][1] >= clipcutoff:
        clipped_len = cigar[-1][1]
        right_clip = cigar[-1][1] * -1
        clipped_frag = read.seq[right_clip:]
        candidate_frag.append(clipped_frag)
    # If the clipped fragment is from Sequencing primer.
    if not read.is_reverse:
        candidate_frag = [x for x in candidate_frag if x not in ME_S7 and x not in S5_ME]
    else:
        candidate_frag = [x for x in candidate_frag if x not in ME_S7_revc and x not in S5_ME_revc]
    return candidate_frag

def ClipInTarget(read, clipcutoff, target):
    # Only clipped on one end of the read is allowed
    candidate_frag = GetClipFrag(read, clipcutoff)

    if len(candidate_frag) == 0:
        return ([], "", 0, 0)

    target_seq_str = str(target)
    target_revcseq_str = str(target.reverse_complement())
    target_seq_len = len(target_seq_str)

    for clipped_frag in candidate_frag:
        clipped_len = len(clipped_frag)
        three_prime_pos = target_seq_len - clipped_len

        plus_hits = [x.start() + 1 for x in re.finditer(clipped_frag, target_seq_str)]
        minus_hits = [x.start() for x in re.finditer(clipped_frag, target_revcseq_str)]
        hits = plus_hits + minus_hits

        if len(plus_hits) == 1 and len(minus_hits) == 0:
            return (hits, "+", clipped_len, three_prime_pos)
        elif len(minus_hits) == 1 and len(plus_hits) == 0:
            return (hits, "-", clipped_len, three_prime_pos)
        else:
            return ([], "", 0, 0)

def FindChimericPair(unpaired_reads, chimera_fragments):
    for readname in unpaired_reads:
        try:
            read1, read2 = unpaired_reads[readname]
        except:
            continue

        if args.candidate_bam:
            candidate_fh.write(read1)
            candidate_fh.write(read2)
        if read1.reference_name == vector_name:
            VectorRead = read1
            HostRead = read2
        elif read2.reference_name == vector_name:
            VectorRead = read2
            HostRead = read1
        else:
            raise Exception("FindChimericPair is processing non-chimeric read pairs.")

        VectorRead_alts = getAltanatives(VectorRead.tags)
        HostRead_alts = getAltanatives(HostRead.tags)
        VectorRead_alts_chrs = [x[0] for x in VectorRead_alts]
        HostRead_alts_chrs = [x[0] for x in HostRead_alts]

        Quality = min(VectorRead.mapq, HostRead.mapq)
        # if HostRead.mapq == 0:
        #     continue
        if vector_name in HostRead_alts_chrs or HostRead.reference_name in VectorRead_alts_chrs:  # if there is host read alternatives on provirus sequence
            continue
        elif len([x for x in VectorRead_alts_chrs if x != vector_name]) > 0:  # if there is vector read alternatives on host genome
            continue
        elif len(VectorRead_alts_chrs) == 1 and vector_name in VectorRead_alts_chrs:
            Quality = HostRead.mapq

        if HostRead.is_reverse != VectorRead.is_reverse:
            InsertOri = "+"
        else:
            InsertOri = "-"
        Chrom = HostRead.reference_name
        ChrStart = HostRead.reference_start + 1
        ChrEnd = HostRead.reference_end
        VectorStart = VectorRead.reference_start + 1
        VectorEnd = VectorRead.reference_end
        '''
        If one of the Read is clipped and the clipped fragment can be found in LTR
        or local sequence within 200 bp of the host read aligned to, we can identify 
        the exact integration loci.
        '''
        ltr_seq = LTR_SEQ if LTR_SEQ else vector_seq[:ltr_len]
        ltr_hits, ltr_strand, clipped_len, ltr_three_prime_pos = ClipInTarget(HostRead, args.LTRClipLen, ltr_seq)

        # Retrieve the local sequence upper or downstream of Host Read aligned position.
        local_seq_st = False
        local_seq_ed = False
        if InsertOri == "+" and VectorRead.cigar[0][0] == 4 and VectorRead.cigar[0][1] >= args.LTRClipLen:
            local_seq_st = HostRead.reference_end
            local_seq_ed = local_seq_st + 200
            host_local_ltr = "LTR5"
        elif InsertOri == "+" and VectorRead.cigar[-1][0] == 4 and VectorRead.cigar[-1][1] >= args.LTRClipLen:
            local_seq_st = HostRead.reference_start - 200
            local_seq_ed = local_seq_st + 200
            host_local_ltr = "LTR3"
        elif InsertOri == "-" and VectorRead.cigar[0][0] == 4 and VectorRead.cigar[0][1] >= args.LTRClipLen:
            local_seq_st = HostRead.reference_start - 200
            local_seq_ed = local_seq_st + 200
            host_local_ltr = "LTR5"
        elif InsertOri == "-" and VectorRead.cigar[-1][0] == 4 and VectorRead.cigar[-1][1] >= args.LTRClipLen:
            local_seq_st = HostRead.reference_end
            local_seq_ed = local_seq_st + 200
            host_local_ltr = "LTR3"

        if local_seq_st and local_seq_ed:
            twoBitToFa_P = Popen("twoBitToFa -seq={chr} -start={start} -end={end} {TwoBit} stdout".format(
                chr=Chrom, start=local_seq_st, end=local_seq_ed, TwoBit=args.Host2bit), shell=True, stdout=PIPE)
            twoBitToFa_P.wait()
            
            try:
                host_local_seq = SeqIO.parse(twoBitToFa_P.stdout, 'fasta').next().seq
                host_local_hits, host_local_strand, clipped_len, host_three_prime_pos = ClipInTarget(VectorRead, args.LTRClipLen, host_local_seq)
            except:
                host_local_hits = []
                host_local_strand = ""
        else:
            host_local_hits = []
            host_local_strand = ""

        LTR = "None"
        if len(ltr_hits) == 1 and ltr_strand == InsertOri and len(host_local_hits) == 0:
            if ltr_hits[0] <= 2:
                LTR = "LTR5"
                ChrStart = HostRead.reference_end
                VectorStart = ltr5_start
                VectorEnd = ltr5_start
            elif ltr_hits[0] >= ltr_three_prime_pos - 2:
                LTR = "LTR3"
                ChrEnd = HostRead.reference_start + 1
                VectorStart = ltr3_end
                VectorEnd = ltr3_end
        elif len(host_local_hits) == 1 and host_local_strand == InsertOri and len(ltr_hits) == 0:
            if host_local_strand == "+":
                ChrStart = local_seq_st + host_local_hits[0] + 1
                ChrEnd = ChrStart
            else:
                ChrStart = 200 - host_local_hits[0] + local_seq_st
                ChrEnd = ChrStart
            if VectorRead.reference_start == 0:
                VectorStart = ltr5_start
                VectorEnd = ltr5_start
                LTR = "LTR5"
            elif VectorRead.reference_end == vector_len:
                VectorStart = ltr3_end
                VectorEnd = ltr3_end
                LTR = "LTR3"
            else:
                ChrStart = HostRead.reference_start + 1
                ChrEnd = HostRead.reference_end

        integration = Integration(Genome=args.genome, Chrom=Chrom, ChrStart=ChrStart, ChrEnd=ChrEnd,
                                  VectorStart=VectorStart, VectorEnd=VectorEnd, LTR=LTR, InsertOri=InsertOri, Quality=Quality)
        chimera = ChimericPair(HostRead=HostRead, VectorRead=VectorRead, integration=integration)
        chimera_fragments.append(chimera)

def FindHostClip(read, clipcutoff, chimera_fragments):
    ltr_seq = LTR_SEQ if LTR_SEQ else vector_seq[:ltr_len]
    hits, strand, clipped_len, three_prime_pos = ClipInTarget(read, clipcutoff, ltr_seq)
    if len(hits) != 1:
        return
    '''
    1. Insertion Orientation.
        Since the read seq in bam file is always in the same orientation with reference sequence, clipped frag strand 
        alone can determine insertion orientation.
        "+" if the clipped fragment is on the positvie strand, "-" if it's on negative
    2. 5' LTR or 3' LTR.
        a. 5' if the integration is at the first base of the sequence, 3' if it's the end
        b. integration site on host should be the end of the aligned position if it's 5' LTR, start position otherwise.
    3. why 3 or three_prime_pos-3
        A few bases (we set 3 at most here) from LTR at the integration site might be able to aligned to host genome 
        together with the host fragment, in this case, we will lost a few bases at the start/end of LTR.
    '''
    if args.candidate_bam:
        candidate_fh.write(read)
    InsertOri = strand

    if hits[0] <= 3:
        LTR = "LTR5"
        ChrStart = read.reference_end
        VectorStart = ltr5_start
        VectorEnd = hits[0] + clipped_len -1
        vector_integration_pos = VectorStart
    elif hits[0] >= three_prime_pos - 3:
        LTR = "LTR3"
        ChrStart = read.reference_start + 1
        VectorStart = hits[0]
        VectorEnd = ltr3_end
        vector_integration_pos = VectorEnd
    else:
        return
    Quality = read.mapq
    ChrEnd = ChrStart
    # VectorEnd = VectorStart
    ChrEnd = ChrStart
    integration = Integration(Genome=args.genome, Chrom=read.reference_name, ChrStart=ChrStart,
                              ChrEnd=ChrEnd, VectorStart=vector_integration_pos, VectorEnd=vector_integration_pos, LTR=LTR, InsertOri=InsertOri, Quality=Quality)
    chimera = ChimericRead(read=read, integration=integration)
    chimera.vector_start = VectorStart
    chimera.vector_end = VectorEnd
    chimera_fragments.append(chimera)

def ReAlignment(fafile, index, chimera_fragments, vector_reads):
    import subprocess
    samfile = fafile + ".sam"
    command = "bwa mem -T {quality} -k {seed} -a -Y  -q {index} {fa} -o {sam}".format(
        index=index, fa=fafile, sam=samfile, quality=args.HostClipLen, seed=args.HostClipLen - 2)
    child = subprocess.Popen(command, shell=True)
    child.wait()
    if child.poll() != 0:
        raise Exception

    re_align = pysam.AlignmentFile(samfile, 'r')
    for clipread in re_align:
        if clipread.mapq == 0:
            continue
        clipread_name = clipread.qname.split(":")
        barcode = clipread_name[0]
        readname = ":".join(clipread_name[1:])
        try:
            read1, read2 = vector_reads.pop(readname)
        except:
            print("{} has more than one non-zero hits in clipped fragment ReAlignment.".format(readname))
            continue
        
        if clipread.is_reverse:
            clipread.seq = str(Seq(clipread.seq).reverse_complement())

        if clipread.seq in read1.seq:
            read = read1
        elif clipread.seq in read2.seq:
            read = read2
        else:
            print(read2.tostring())
            print(read1.tostring())
            sys.exit()
        clipped_loc = read.seq.index(clipread.seq)
        InsertOri = "+" if read.is_reverse == clipread.is_reverse else "-"
        if clipped_loc <= 3:
            VectorStart = read.reference_start + 1
            ChrStart = clipread.reference_end
        else:
            VectorStart = read.reference_end
            ChrStart = clipread.reference_start + 1

        LTR = "None"
        if (InsertOri == "+" and abs(VectorStart - ltr5_start) <= 3) or (InsertOri == "-" and abs(VectorStart - ltr3_end) <= 3):
            LTR = "LTR5"
        elif (InsertOri == "+" and abs(VectorStart - ltr3_end) <= 3) or (InsertOri == "-" and abs(VectorStart - ltr5_start) <= 3):
            LTR = "LTR3"
        else:
            continue
        Chrom = clipread.reference_name
        ChrEnd = ChrStart
        VectorEnd = VectorStart
        integration = Integration(Genome=args.genome, Chrom=Chrom, ChrStart=ChrStart, ChrEnd=ChrEnd, VectorStart=VectorStart,
                                  VectorEnd=VectorEnd, LTR=LTR, InsertOri=InsertOri, CellNumber=1, CellBarcodes=[barcode],
                                  ReadInCell=[1], ReadNumber=1, Quality=clipread.mapq)
        chimera = ChimericRead(read=read, integration=integration)
        chimera.reference_name = clipread.reference_name
        chimera.reference_start = clipread.reference_start +1
        chimera.reference_end = clipread.reference_end
        chimera.vector_start = read1.reference_start +1
        chimera.vector_end = read2.reference_end
        chimera_fragments.append(chimera)

def CorrectLTR(read1, read2):
    read1_alts = getAltanatives(read1.tags)
    read2_alts = getAltanatives(read2.tags)
    insert_size = abs(read1.template_length)

    read1_vector_alts = [x for x in read1_alts if x[0] == vector_name]
    read2_vector_alts = [x for x in read2_alts if x[0] == vector_name]

    # Here we only allow clipped alignment on one read, and full Alignment on the other. 3 bases bias allowed at the clipped site.
    if len(read1_vector_alts) == 1 and len(read2_vector_alts) == 1 and insert_size > read1.qlen:
        if read2.cigar[-1][0] == 4 and "S" not in read1.cigarstring and abs(read2.reference_end - ltr5_end) <= 3:
            read2.reference_start = int(read2_vector_alts[0][1].lstrip("[-+]")) - 1
            read1.reference_start = int(read1_vector_alts[0][1].lstrip("[-+]")) - 1
        elif read1.cigar[0][0] == 4 and "S" not in read2.cigarstring and abs(read1.reference_start - ltr3_start) <= 3:
            read1.reference_start = ltr5_start
            read2.reference_start = int(read2_vector_alts[0][1].lstrip("[-+]")) - 1

        read1.next_reference_start = read2.reference_start
        read2.next_reference_start = read1.reference_start

    return read1, read2


def ProcessVectorAlignments(vector_reads, fa_fh, vector_fragments, clipcutoff):
    for readname in vector_reads:
        try:
            read1, read2 = vector_reads[readname]
        except:
            continue
        '''
        Since there are two identical LTRs on each end of the sequence, reads that clipped at beginning of LTR5 
        and end of LTR3 can also be alternatively aligned to the other. If the read clipped aligned to the end 
        of LTR5 or start of LTR3, and both reads can be alternatively aligned to the other LTR, then the altertive 
        alignment will be chosen as primary.
        '''
        if read1.is_unmapped or read2.is_unmapped:
            continue
        if read1.reference_start > read2.reference_start:
            read1, read2 = read2, read1

        if args.candidate_bam:
            candidate_fh.write(read1)
            candidate_fh.write(read2)
        read1, read2 = CorrectLTR(read1, read2)
        # Only clipped on one end of the read is allowed
        read1_candidate_frag = GetClipFrag(read1, clipcutoff)
        read2_candidate_frag = GetClipFrag(read2, clipcutoff)

        barcode = get_barcode(read1)
        if len(read1_candidate_frag) == 1 and len(read2_candidate_frag) == 0:
            read_fa = ">" + barcode + ":" + read1.qname + "\n" + read1_candidate_frag[0] + "\n"
            fa_fh.write(read_fa)
        elif len(read1_candidate_frag) == 0 and len(read2_candidate_frag) == 1:
            read_fa = ">" + barcode + ":" + read2.qname + "\n" + read2_candidate_frag[0] + "\n"
            fa_fh.write(read_fa)

        vector_pair = VectorFragment(first=read1, second=read2)
        vector_fragments.append(vector_pair)
    return


def ParseBam(bamfile, unpaired_reads, chimera_fragments, vector_reads):
    align = pysam.AlignmentFile(bamfile, 'rb', threads=20)
    for read in align:
        readname = read.qname
        if read.flag & 2 and read.reference_name != vector_name and "S" in read.cigarstring:
            FindHostClip(read, args.LTRClipLen, chimera_fragments)
        elif read.reference_name == vector_name and read.next_reference_name == vector_name:
            vector_reads[readname].append(read)
        elif (not read.flag & 14) and (read.reference_name == vector_name or read.next_reference_name == vector_name):
            unpaired_reads[readname].append(read)
    return


def main():
    chimera_fragments = []
    vector_fragments = []
    vector_reads = defaultdict(list)
    unpaired_reads = defaultdict(list)

    t1 = time.time()
    print("Start Parsing Bam", file=sys.stderr)
    ParseBam(args.bamfile, unpaired_reads, chimera_fragments, vector_reads)
    print("Finish Parsing Bam {}".format(time.time() - t1), file=sys.stderr)
    FindChimericPair(unpaired_reads, chimera_fragments)
    print("Finish ChimericPair {}".format(time.time() - t1), file=sys.stderr)
    fapath = os.path.join(args.tempdir, "Clipped_fragment.fa")
    fa_fh = open(fapath, 'w')
    ProcessVectorAlignments(vector_reads, fa_fh, vector_fragments, args.HostClipLen)
    fa_fh.close()
    if os.path.getsize(fapath) != 0:
        ReAlignment(fapath, args.HostIndex, chimera_fragments, vector_reads)
    print("Finish VectorFragment {}".format(time.time() - t1), file=sys.stderr)

    vector_coverage = os.path.join(args.outdir, "Vector_coverage.csv")
    report_file = os.path.join(args.outdir, "Cell_summary.txt")
    
    chimera_counter = ChimeraSummary(chimera_fragments, args.outdir, args.gbdb, args.quality)
    vector_counter = VectorSummary(vector_fragments, vector_coverage, args.outdir)
    summary_counter = defaultdict(dict)
    for barcode in chimera_counter:
        summary_counter[barcode] = chimera_counter[barcode]

    for barcode in vector_counter:
        if barcode in summary_counter:
            summary_counter[barcode].update(vector_counter[barcode])
        else:
            summary_counter[barcode] = vector_counter[barcode]

    summary_counter_df = pd.DataFrame.from_dict(summary_counter, orient = 'index')
    summary_counter_df.index.name = 'CellBarcode'
    summary_counter_df.to_csv(report_file, sep="\t", na_rep=0)
    print("Finish pipeline {}".format(time.time() - t1), file=sys.stderr)

if __name__ == '__main__':
    main()



from setuptools import setup, find_packages

with open("README.md", "r") as fh:
    long_description = fh.read()



setup(
    name="epiVIA",
    version="1.1.2",
    author="Wenliang Wang",
    author_email="wangwl.me@gmail.com",
    description="Virial Integration Analysis with epigenetic data",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/wangwl/epiVIA",
    package_dir={'': 'src'},
    packages=find_packages('src'),
    install_requires=['pysam', 'pandas', 'biopython', 'requests', 'pybedtools'],
    entry_points={'console_scripts': ['epiVIA = epiVIA.chimera:main']},
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
    ],
)



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# @Date    : 2019-05-15 22:58:29
# @Author  : Your Name (you@example.org)
# @Link    : http://example.org
# @Version : $Id$

import os
import sys
import re
import json
import yaml
import argparse
import requests
from subprocess import Popen, PIPE

from epiVIA.utils import bb_conf, get_abs_path

UCSC_API = "https://api.genome.ucsc.edu/"

class UCSCTrack(object):
	"""docstring for UCSCTrack"""
	def __init__(self, **kwargs):
		super(UCSCTrack, self).__init__()


# @pysnooper.snoop()
def run_bigBedToBed(track, genome, chrom=None, start=None, end=None, columns=None, gbdb="http://hgdownload.soe.ucsc.edu/gbdb/"):
	try:
		file_path = "{gbdb}/{genome}/{bb_path}".format(gbdb=gbdb, genome=genome, bb_path=bb_conf[track])
		cmd = "{bigBedToBed_path} {bb_path} stdout".format(bigBedToBed_path=get_abs_path('bigBedToBed'), bb_path=file_path)
		if chrom:
			cmd += " -chrom={chrom}".format(chrom=chrom)
		if start:
			cmd += " -start={start}".format(start=start)
		if end:
			cmd += " -end={end}".format(end=end)

		child = Popen(cmd, shell=True, stdout=PIPE, universal_newlines=True)
		child.wait()
		# if child.poll() != 0:
		# 	raise Exception
		output = []
		for line in child.stdout.readlines():
			row = str(line).rstrip().split("\t")
			record = []
			if len(columns) > 0:
				for x in columns:
					record.append(row[x])
			else:
				record = row
			output.append(record)
		return output
	except:
		print("fail to annotate {}, continue anyway.".format(track))
		return []
	## bigBedToBed subprocess command line
	
	

#@pysnooper.snoop()
def find_ucsctrack(track, genome, chrom=None, start=None, end=None, keys=[]):
	action = "getData"
	datatype = "track"
	track_name = track

	url = UCSC_API + "{action}/{datatype}?genome={genome};track={track}".format(action = action, datatype=datatype, genome=genome, track=track_name)
	if chrom:
		url += ";chrom={chrom}".format(chrom=chrom)
	if start:
		url += ";start={start}".format(start=start)
	if end:
		url += ";end={end}".format(end=end)
	request = requests.get(url)

	# if request.status_code != 200:
	# 	print url
	if request.status_code == 200 and request.headers['Content-Type'] == 'application/json':
		content = yaml.safe_load(request.text)
		items = content[track_name]
		output = []
		for record in items:
			line = []
			if len(keys) > 0:
				for key in keys:
					line.append(record[key])
				output.append(line)
			else:
				output.append(record)
		return output
	else:
		print("fail to annotate {} with UCSC API, continue anyway.".format(track))
		return []

def find_trackhub():
	pass

def chunk_image(imgdir, **kwargs):
	IMAGE_API = "http://genome.ucsc.edu/cgi-bin/hgRenderTracks?"
	request = requests.get(IMAGE_API, params=kwargs)
	if request.status_code != 200:
		raise Exception

	image_file = os.path.join(imgdir, kwargs['position']+".png")
	image_fh = open(image_file, 'w')
	image_fh.write(request.content)
	image_fh.close()

def main():
	parser = argparse.ArgumentParser(description='annotate chimeric region with TE and gene features')
	parser.add_argument("--bedfile", help="a bedfile with the chimeric region")
	parser.add_argument("track")
	parser.add_argument("--track_keys", nargs="+")
	parser.add_argument("--bedcolumns", nargs="+", type=int)
	parser.add_argument("--sequence")
	parser.add_argument("--hubUrl")
	parser.add_argument("--genome")
	parser.add_argument("--chrom")
	parser.add_argument("--start")
	parser.add_argument("--end")
	parser.add_argument("--imgdir")
	parser.add_argument("--gbdb", default="http://hgdownload.soe.ucsc.edu/gbdb/")
	args = parser.parse_args()
	if args.track_keys:
		items = find_ucsctrack(args.track, args.genome, args.chrom, args.start, args.end,  args.track_keys)
	else:
		items = find_ucsctrack(args.track, args.genome, args.chrom, args.start, args.end, args.bedcolumns)
	for x in items:
		print("\t".join(x))

if __name__ == '__main__':
	main()


#!/usr/bin/env python
# -*- coding: utf-8 -*-
# @Date    : 2019-05-10 14:41:41
# @Author  : Your Name (you@example.org)
# @Link    : http://example.org
# @Version : $Id$

import os
import sys
import pybedtools
from epiVIA.ucsctracks import find_ucsctrack, find_trackhub, chunk_image, run_bigBedToBed
'''

'''

class Integration(object):
    """docstring for integration
    # Integration Attributes: Chrom ChrStart ChrEnd HotLevel VectorStart VectorEnd LTR  InsertOri   InsertGene  \
    GeneOri  ExonIntron TE TEFamily  TEClass NearestGene    NearestGeneDist  CellNumber  CellBarcodes
    # Example: hg38 chr7    114399221   114399221   5752    5752    LTR3    +   FOXP2   +   intron  None    None \
    Alu SINE    1   ATGATGCACAATAGCA    1   1   HostRead
    """

    def __init__(
        self,
        Genome='hg38',
        Chrom=None,
        ChrStart=None,
        ChrEnd=None,
        HotLevel=None,
        VectorStart=None,
        VectorEnd=None,
        LTR=None,
        InsertOri=None,
        InsertGene=None,
        GeneOri=None,
        ExonIntron=None,
        Enhancer=None,
        Elite=None,
        TEFamily=None,
        TEClass=None,
        UpstreamGene=None,
        UpstreamGeneDist=None,
        CellBarcodes=None,
        ReadInCell=None,
        CellNumber=None,
        ReadNumber=None,
        ChimeraSource=None,
        Quality = None
    ):
        super(Integration, self).__init__()
        self.Genome = Genome
        self.Chrom = Chrom
        self.ChrStart = ChrStart
        self.ChrEnd = ChrEnd
        self.VectorStart = VectorStart
        self.VectorEnd = VectorEnd
        self.LTR = LTR
        self.InsertOri = InsertOri
        self.InsertGene = InsertGene
        self.GeneOri = GeneOri
        self.ExonIntron = ExonIntron
        self.Enhancer = Enhancer
        self.Elite = Elite
        self.TEFamily = TEFamily
        self.TEClass = TEClass
        self.UpstreamGene = UpstreamGene
        self.UpstreamGeneDist = UpstreamGeneDist
        self.CellBarcodes = CellBarcodes
        self.ReadInCell = ReadInCell
        self.CellNumber = CellNumber
        self.ReadNumber = ReadNumber
        self.ChimeraSource = ChimeraSource
        self.Quality = Quality

    def annotate_TE(self):
        ChrEnd = self.ChrEnd + 1 if self.ChrEnd == self.ChrStart else self.ChrEnd
        results = find_ucsctrack('rmsk', self.Genome, self.Chrom, self.ChrStart, ChrEnd, [
                                 'repClass', 'repFamily'])

        try:
            repClass, repFamily = results[0]
        except:
            repClass = "None"
            repFamily = "None"

        self.TEFamily = repFamily
        self.TEClass = repClass
        return self

    def annotate_Gene(self, gbdb="http://hgdownload.soe.ucsc.edu/gbdb/"):
        ChrEnd = self.ChrEnd + 1 if self.ChrEnd == self.ChrStart else self.ChrEnd
        results = []
        # try:
        #     results = find_ucsctrack('knownGene', self.Genome, self.Chrom, self.ChrStart, ChrEnd, ['txStart', 'txEnd', 'strand', 'exonStarts', 'exonEnds', 'name'])
        #     # print(results)
        #     if len(results) > 0:
        #         GeneSt, GeneEd, GeneOri, Exon_Starts, Exon_Ends, GeneSymble = results[0]
        #         Exon_Starts = [int(x) for x in Exon_Starts.rstrip(",").split(",")]
        #         Exon_Ends = [int(x) for x in Exon_Ends.rstrip(",").split(",")]
        # except:
        results = run_bigBedToBed('knownGene', self.Genome, self.Chrom, self.ChrStart, ChrEnd, [1, 2, 5, 10, 11, 17], gbdb)
        if len(results) > 0:
            GeneSt, GeneEd, GeneOri, ExonLen, ExonSt, GeneSymble = results[0]
            GeneSt = int(GeneSt)
            # print(results[0])
            Exon_Lengths = [int(x) for x in ExonLen.rstrip(",").split(",")]
            Exon_Starts = [GeneSt + int(ExonSt.rstrip(",").split(",")[x]) for x in range(len(Exon_Lengths))]
            Exon_Ends = [Exon_Starts[x] + Exon_Lengths[x] for x in range(len(Exon_Lengths))]
            

            for x in range(len(Exon_Starts)):
                if self.ChrStart >= Exon_Starts[x] and self.ChrStart <= Exon_Ends[x]:
                    ExonIntron = "exon"
                elif x + 1 < len(Exon_Starts):
                    if self.ChrStart > Exon_Ends[x] and self.ChrStart <= Exon_Starts[x + 1]:
                        ExonIntron = "intron"
                    elif (self.ChrStart >= Exon_Starts[x] and self.ChrStart <= Exon_Ends[x]) and (self.ChrEnd >= Exon_Starts[x + 1] and self.ChrEnd <= Exon_Ends[x + 1]):
                        ExonIntron = 'junction'
            try:
                ExonIntron
            except:
                ExonIntron = "None"
        else:
            ExonIntron = "None"
            GeneSymble = "None"
            GeneOri = "None"

        self.ExonIntron = ExonIntron
        self.InsertGene = GeneSymble
        self.GeneOri = GeneOri
        return self

    def annotate_Enhancer(self, gbdb="http://hgdownload.soe.ucsc.edu/gbdb/"):
        ChrEnd = self.ChrEnd + 1 if self.ChrEnd == self.ChrStart else self.ChrEnd
        # try:
        #     results = find_ucsctrack('geneHancerRegElements', self.Genome, self.Chrom, self.ChrStart, ChrEnd, ['elementType', 'eliteness'])
        # except:
        results = run_bigBedToBed('geneHancerRegElements', self.Genome, self.Chrom, self.ChrStart, ChrEnd, [-2, -1], gbdb)
        
        try:
            elementType, eliteness = results[0]
        except:
            elementType = "None"
            eliteness = "None"

        self.Enhancer = elementType
        self.Elite = eliteness
        return self

    def annotate_UCSC_track(self, track):
        pass

    def nearest_gene(self):
        ChrEnd = self.ChrEnd + 1 if self.ChrEnd == self.ChrStart else self.ChrEnd
        integration_site_bed = pybedtools.BedTool("{chrom} {start} {end}".format(chrom=self.Chrom, start=self.ChrStart, end=ChrEnd), from_string=True)
        chrom_genes = find_ucsctrack('knownGene', self.Genome, self.Chrom)
        chrom_genes_list = []
        for gene in chrom_genes:
            '''
                    The result from find_ucsctrack can be either a list of dict or list of list, based on retrieve from bigBed or API.
                    1. Content of each gene as a dict:
                    {'alignID': 'uc033dmk.2',
                      'cdsEnd': 1670,
                      'cdsStart': 1670,
                      'chrom': 'chrM',
                      'exonCount': 1,
                      'exonEnds': '3229,',
                      'exonStarts': '1670,',
                      'name': 'ENST00000387347.2',
                      'proteinID': '',
                      'strand': '+',
                      'txEnd': 3229,
                      'txStart': 1670}
                    2. Content of each gene as a list:
                    [chrM, 1670, 3229, ENST00000387347.2, 3, +, 1670, 1670, 789624, 1, "1559,", "0,", ENST00000387347.2, none, none, "-1,", , MT-RNR2, , ]
                    '''
            if isinstance(gene, dict):
                chrom_genes_list.append("{}\t{}\t{}\t{}\t{}".format(gene['chrom'], gene['txStart'], gene['txEnd'], gene['name'], gene['strand']))
            if isinstance(gene, list):
                chrom_genes_list.append("{}\t{}\t{}\t{}\t{}".format(gene[0], gene[1], gene[2], gene[3], gene[5]))
        chrom_genes_string = "\n".join(chrom_genes_list)
        chrom_genes_bed = pybedtools.BedTool(chrom_genes_string, from_string=True)
        chrom_genes_bed = chrom_genes_bed.sort()
        upstream_closest = integration_site_bed.closest(chrom_genes_bed, d=True, D='ref', io=True, id=True)
        downstream_closest = integration_site_bed.closest(chrom_genes_bed, d=True, D='ref', io=True, iu=True)
        # The closest BedTool object will be in the format of:
        # chrM  1000    1300    chrM    576 647 ENST00000387314.1   +   -354
        for gene in upstream_closest:
            self.UpstreamGeneDist = gene[-1]
            self.UpstreamGeneOri = gene[-2]
            self.UpstreamGene = gene[-3]
        for gene in downstream_closest:
            self.DownstreamGeneDist = gene[-1]
            self.DownstreamGeneOri = gene[-2]
            self.DownstreamGene = gene[-3]
        return self

    def visualize_UCSC_track(self, imgdir):
        img_start = str(self.ChrStart - 10000)
        img_end = str(self.ChrEnd + 10000)
        position = "{Chrom}:{start}-{end}".format(
            Chrom=self.Chrom, start=img_start, end=img_end)
        params = {'db': self.Genome, 'position': position}
        chunk_image(params, imgdir=imgdir)

    def distance_to_known(self, konwnsites):
        pass

    def perturbation(self):
        pass

    def tostring(self):
        try:
            print_string = "\t".join([self.Genome, self.Chrom, str(self.ChrStart), str(
                self.ChrEnd), str(self.VectorStart), str(self.VectorEnd), self.LTR, self.InsertOri])
            header = "Genome\tChrom\tChrStart\tChrEnd\tVectorStart\tVectorEnd\tLTR\tInsertOri"
        except:
            print(self.Genome, self.Chrom, str(self.ChrStart), str(self.ChrEnd), str(self.VectorStart), str(self.VectorEnd), self.LTR, self.InsertOri)

        try:
            print_string += "\t" + self.InsertGene + \
                "\t" + self.GeneOri + "\t" + self.ExonIntron
            header += "\tInsertGene\tGeneOri\tExonIntron"
        except:
            print("Gene not annoated", file=sys.stderr)

        # try:
        #     print_string += "\t" + self.Enhancer + "\t" + self.Elite
        #     header += "\tEnhancer\tElite"
        # except:
        #     print("Enhancer not annoated", file=sys.stderr)

        try:
            print_string += "\t" + self.TEFamily + "\t" + self.TEClass
            header += "\tTEFamily\tTEClass"
        except:
            print("TE not annoated", file=sys.stderr)

        # try:
        #     print_string += "\t" + self.UpstreamGene + "\t" + \
        #         self.UpstreamGeneOri + "\t" + self.UpstreamGeneDist
        #     header += "\tUpstreamGene\tUpstreamGeneOri\tUpstreamGeneDist"
        # except:
        #     print("UpstreamGene not annoated", file=sys.stderr)

        # try:
        #     print_string += "\t" + self.DownstreamGene + "\t" + \
        #         self.DownstreamGeneOri + "\t" + self.DownstreamGeneDist
        #     header += "\tDownstreamGene\tDownstreamGeneOri\tDownstreamGeneDist"
        # except:
        #     print("DownstreamGene not annoated", file=sys.stderr)

        try:
            print_string += "\t" + str(self.CellNumber) + "\t" + ",".join(self.CellBarcodes) + \
                "\t" + ",".join([str(x) for x in self.ReadInCell]
                                ) + "\t" + str(self.ReadNumber)
            header += "\tCellNumber\tCellBarcodes\tReadInCell\tReadNumber"
        except:
            print("Cell and Read not calculated", file=sys.stderr)

        try:
            print_string += "\t" + ",".join(self.ChimeraSource)
            header += "\tChimeraSource"
        except:
            print("ChimeraSource not defined", file=sys.stderr)

        try:
            print_string += "\t{}".format(self.Quality)
            header += "\tQuality"
        except:
            pass

        print_string += "\n"
        header += "\n"
        return print_string, header


class HotSpot(object):
    """docstring for HotSpot"""

    def __init__(self, arg):
        super(HotSpot, self).__init__()
        self.arg = arg
